library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- revision of CFX-4 output code

entity testVGA2 is
 
    port
	(
 -- VGA controller parts
		clk       		: in std_logic;
		ma        		: out std_logic_vector(14 downto 0);	-- Up to 32k Video ram available
		ra        		: out std_logic_vector(3 downto 0);		-- Up to 16 row high characters
		blank      		: out std_logic;		-- Shift register output enable
		hsync     		: out std_logic;		-- VGA Syncs
		vsync     		: out std_logic;	
		shld      		: out std_logic;		-- Latch pixel data direct from ram
		Colourhold		: out std_logic;		--	Latch temp colour register from ram
		Bypass			: out std_logic;		-- switch pixel data between char rom (A pins) and direct read from VRAM (B pins)
		Option			: in std_logic_vector(0 downto 0);		-- data from HC273 Mode Control register
														-- bit 0 screen type - Text (L) or Graphics (H)
														-- bits 4-7 are border colour
		Tmode				: out std_logic;		-- Shift register mode selection 0 text/ 1 graphics
		PhiV				: out std_logic;		-- Synchronised PHI output for a 65C02 with Phi low during CPLD access time.
		Int4				: out std_logic;		-- Frame end interrupt
		IRQack			: in std_logic;		-- interrupt acknowlwdge from the CPU
		border			: out std_logic;		-- border area colour select
		-- memory access 		
		ramrq				: in std_logic;		-- External memory decoder has identified a request to access the VRAM		
		reset     		: in std_logic;		-- Global reset
		CPUread     	: in std_logic;		-- CPU requesting read access
		CPUWrite   		: in std_logic;		-- CPU requesting write acess
		
		ramcs		  		: out std_logic;		-- Enable the video ram
		Vread				: out std_logic;		-- read access of the Vram
		Vwrite			: out std_logic;		-- write to Vram
		Vrom				: out std_logic;		-- enable for the character rom - not used in graphic modes
		CPUaddrlatch	: out std_logic;		-- Latch the CPU address for ram access when free
		CPUreadlatch	: out std_logic;		-- Latch the VRAM data for CPU access when ready
		CPUwritelatch	: out std_logic;		-- Latch the CPU data for ram write when free
		CPUreadOE		: out std_logic;		-- OE for the read latch, open when the data is ready, close whwn CPU drops/RD
		CPUWriteOE		: out	std_logic;		-- OR for the write latch. only open when the CPU has access to the VRAM
		
		AddrSource		: out std_logic		-- Switch for the 4HC157's presenting the address to the Vramm
		
		) ;

end entity;
-- revised line Back Porch 48, display 640, front porch 16, sysnc 96
architecture behavioral of testVGA2 is
constant H_Display:  integer := 48;
Constant H_porch:		integer := 688;
constant H_Sync:     integer := 704;
constant H_Total:    integer := 800;

signal H_Count:      integer range 0 to H_Total;
signal H_Blank:      std_logic := '0';

-- 25 rows of 16 scan lines = 400 lines, then  border, 12 porch, 2 vertical sync then final 36 porch
constant V_border:	integer := 400;
constant V_Porch:    integer := 440;
constant V_Sync:     integer := 450;
constant V_SyncEnd:  integer := 452;
constant V_topBorder:integer := 488;
constant V_Total:    integer := 525;

signal V_Count:      integer range 0 to V_Total;
signal V_Blank:      std_logic := '0';

constant CharHeight: integer := 16-1;
constant RamSize:    integer := 16384-1;

signal MemPointer:   integer range 0 to RamSize;
signal C_Count:      integer range 0 to RamSize;
signal ScanLine:     integer range 0 to CharHeight;

signal sequence:		integer range 0 to 7 ;
signal mode:			std_logic := '0' ;


begin

tmode <= mode ;

-- RAM access signals need both a chip select and a read/write signal

CPUreadOE <='0' when ramrq='0' AND CPURead='0' else '1' ;  -- Make ram data avaialabe for as long as the CPU needs it

-- VDP interrupt low for blanking time in final displayed screen line
-- interrupt acknowledge pulse from CPU only lasts 1/2 a CPU cycle so has to have priority
VDPinterrupt:
Process(clk)
begin
	if rising_edge(clk) then
		if IRQack = '0' then 
			int4 <= '1' ;
		elsif  (v_count = v_border) and (h_count<32) then
			int4 <= '0' ;
		else
			int4 <= '1' ;
		end if; 
	end if;
end process;

-- CPU clock goes high on cycles 0|1 4|5  when CPU has access
-- can simplify that to a 4 step sequence
-- CPU clock is low dduring the cycles where the CPLD has access to the memory
CPUMemoryAccess:
Process(clk,h_count)
begin
	if rising_edge(clk) then
		case (h_count mod 4) is
		when 0 => 
			phiv <= '1' ;	
			addrSource <= '0' ;			-- CPU address is on the A pinsallow them to settle in the first phase	
		when 1 => 
			phiv <= '1' ;
			ramcs <= ramrq ;
			vread <= CPUread ;
			vwrite <= CPUwrite ;
			CPUaddrLatch <='0';   		-- lock the CPU address latch to provide some hold time
			CPUwriteLatch <= '0' ;		-- CPU access slot the databus write latch closed to hold the data
			cpuReadLatch <= '1' ;
			if  CPUwrite ='0' and ramrq='0' then
				CPUwriteOE   <= '0' ;	-- data from the write latch onto the vram data bus
			end if;
		when 2 =>
			phiv <= '0' ;
			addrSource <= '1' ;			-- CPLD address is on the B pins
			CPUwriteOE <='1';				-- turn of the write data if it was active
			CPUaddrLatch <= '1' ;  		-- open the address latch now were in CPLD access
			CPUwriteLatch <= '1' ;		-- CPU access slot complete the data latch can be released
			cpuReadLatch <= '0' ;		-- Read latch locks to provide some hold time
			-- CPLD has access now
			vread  <= '0' ;				-- the CPLD only reads, and never writes to memory
			vwrite <= '1' ;
			ramcs <= '0' ;
		when 3 =>
			phiv <= '0' ;
		when others =>
	end case;
	end if;
end process;

-- CPLD reads one byte to the external latch in cycles 2&3 and reads another byte in cycle 6&7
-- both bytes are transferred to the output CPLD in cycle 7
CPLDmemoryAccess:
Process(clk,h_count)
begin
	if rising_edge(clk) then
	   if (h_count = 0) then
			c_count <= memPointer ;
		end if;
		case h_count mod 8 is
		when 0 =>
			-- complete latching process.
			SHLD <= '1' ;
		when 1 =>
			-- setup the address the cycle before it's switched over by the 157's
			if mode='0' then
				MA(10 downto 0) <= std_logic_vector(to_unsigned(C_Count, 11));
				MA(11) <='1';					-- colour data is the top 2k
				ma(14 downto 12) <= "111";	-- text mode screen occupies the top of vram
			else
			   ma(14 downto 1) <= std_logic_vector(to_unsigned(C_count, 14));
				ma(0) <='0' ;				-- in graphics mode the "colour" register is treated as data
			end if;
		when 2 =>
			-- with the address set up, the counter can be incremented but only in the active area and after the first pixels are addressed
			if (h_count>H_display) and (h_count< H_sync) then
				c_count <= C_count + 1 ;
			end if;
		when 3 =>
			colourHold <= '0'	;			-- holding register clock also needs to go low before latching on the next cycle
		when 4 =>
			colourHold <= '1'	;			-- transfer the colour data to the temporary register on the rising edge
		when 5 =>
			if mode='0' then
				MA(11) <= '0' ;			-- flip A11 of VRAM address to get the character data
			else
				ma(0) <= '1' ;				-- graphics mode, get the next byte 
			end if;
		when 6 =>
		when 7 =>	
			SHLD <= '0' ;					-- tell the output CPLD to latch
		when others =>
		end case;	
	end if;
end process;

-- global blanking signal
blank <= '1' when (
(V_Blank='1') OR 
(H_Blank='1')
) else '0' ;

-- border colour select
border <= '1' when (h_count<h_porch) and (((v_count>=v_border) and (v_count<v_porch)) or (v_count>=v_topborder)) else '0' ;

-- Option bit 1 determines graphics or character mode
-- when high the character rom is bypassed, 
-- Possibly have this driven by the register, with the CPLD just monitoring ??

bypass <= '0' when mode='0' else '1' ;

vrom <= '1' when mode='1' else '0' ;    
-- Horizontal Timing

-- Master horizontal counter, in character positions.
-- H_LineEnd is set for the last character position in the row, when this falls
-- again at the start of the next row, that triggers the vertical display actions 

HorizontalCount:
process(Clk, reset)
begin
    if (reset = '0') then
        H_Count <= 0;
    elsif rising_edge(Clk) then
        if H_Count = (H_Total - 1) then
            H_Count <= 0;
        else
             H_Count <= H_Count + 1;
        end if;
    end if;
end process;

-- The display is "active" up to the start of the front porch, after that is
-- either sync or porch until the next row starts, for the the output multiplexer
-- all non displayed area is the same, so only a generic "blank" is needed.

HorizontalDisplay:
process(Clk)
begin
    if rising_edge(Clk) then
        if (h_count>1) and (H_Count <= H_Porch+1) then
            H_Blank <= '0';
        else
            H_Blank <= '1';
        end if;
    end if;
end process;

-- Horizontal sync is active in the blank part of the display
-- It has it's own output, so won't interfere with the blanking signal
-- Triggered by the master clock, not the character clock to get an
-- immedialte update without waitng for the complete dcharacter cell
-- in 640 x 400 mode sysnc is negative

HorizontalSync:
process(clk)
begin
    if rising_edge(clk) then
        if (H_Count >= H_Sync) then
            HSYNC <= '0';
        else
            HSYNC <= '1';
        end if;
    end if;
end process;

--
-- Vertical Timing

-- Master vertical counter, counting in rows triggered by the line end pulse

VerticalCount:
process (clk, reset)
begin
    if (reset = '0') then
			V_Count <= 0;
    elsif rising_edge(clk) then
		if (h_count = h_sync) then
        if (V_Count = (V_Total-1)) then
            V_Count <= 0;
        else
            V_Count <= V_Count + 1;
        end if;
       end if;
    end if;
end process;

-- Similar to the horizontal counter, the display is active from row 0, up to
-- the start of the porch area, where the blanking signal is set until the
-- counter resets 

VerticalDisplay:
process(clk)
begin
    if rising_edge(clk) then
        if (V_Count < V_Border) then
            V_Blank <= '0';
        else
            V_Blank <= '1';
        end if;
    end if;
end process;

-- Vertical sync is active for 2 full rows on the middle of the blanking period
-- a good time to check if the video mode has chnaged
--- Sync on 640 x 400 mode is positive
--- Sync on 640 x 480 mode is negative

VerticalSync:
process(clk)
begin
    if rising_edge(clk) then
        if ((V_Count>= V_Sync) and (V_Count < V_SyncEnd) ) then
            VSYNC <= '0';
				mode				<= option(0);
      else
            VSYNC <= '1';
        end if;
    end if;
end process;

-- Main Scanline Counter

--- The 16 row high characters need a 4 bit counter with no spare rows
-- no blank entries required in the character rom to maintain spacing
-- board built to allow 20 high characters so RA does have 5 pins on the rom
-- 16 high characters needs 16 blanks

ScanLineCounter:
process(clk, reset)
begin
	if rising_edge(clk) then
			scanline <= v_count mod 16 ;
	end if;
end process;
RA <= std_logic_vector(to_unsigned(ScanLine, RA'length));

MemoryCounter:
process(clk, reset)
begin
	if rising_edge(clk) then
		if v_count=(v_total-1) then
			MemPointer <= 0 ;
		elsif	h_count=h_sync then
			if mode ='0' then
			-- text mode the memory address will "stick" until all 16 rows of the character are output
				if (ScanLine = CharHeight)  then
						MemPointer <= MemPointer + 80;	
				end if;
			else
			-- graphics mode, memory address will increment other line
				if v_count mod 2 = 1 then
					MemPointer <= MemPointer + 80;			 
				end if;
			end if;
		end if;
	end if;
end process;

end behavioral;